@@ -23,6 +23,18 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  * MA 02110-1301, USA
+ *
+ *
+ *
+ * 
+ * - First version had hardcoded offset, with a SHA1 sum check
+ *   (Ronald)
+ *   
+ * - Second version, multiple sha1sum were handled with their offset
+ *   (Étienne)
+ *   
+ * - Third version, we search for a preambule in driver and extract
+ *   the firmware (Étienne, idea from Jeroen van der Vegt)
  */
 
 #ifdef HAVE_CONFIG_H
@@ -36,135 +48,6 @@
 #include <glib/gi18n.h>
 #include <glib/gstdio.h>
 
-/*
- * Foreach file identified by their sha1sum, we store the offset where
- * the firmware begin in the file.
- *
- * If a file does not work, you may need to add an entry here. To find
- * the offset in the driver file, use a hexadecimal editor (e.g. ghex)
- * and search for "USBSend". You will find data like this :
- *
- * "start USBSend #4 returned 0x%x.....X.....4.~.........%\"
- *                                                  ^
- *
- * Each period means 0x00. The firmware start at the ^ sign
- * position. Read the offset and store in in this array. The tool must
- * output the sha1sum of the file, use this to fill the .sha1sum
- * field.
- *
- */
-
-struct offset {
-	gchar *desc;
-	unsigned char sha1sum[20];
-	long offset;
-};
-
-static const struct offset offsets[] = {
-	{ "Mac OS X.6 MacBook 1,1",
-	  .sha1sum	= { 0x32, 0xcc, 0xa3, 0x17, 0xae, 0x54, 0xb5, 0x1c,
-			    0x4c, 0xfa, 0x66, 0x4d, 0x0d, 0xa0, 0x47, 0xe9,
-			    0x11, 0x01, 0xf9, 0x1e },
-	  .offset	= 0x20D8,
-	},
-	{ "Mac OS X.6 MacBook 5,2",
-	  .sha1sum	= { 0x06, 0x2d, 0x20, 0xde, 0x72, 0xb9, 0x6a, 0x24,
-			    0xb7, 0x41, 0xc9, 0x51, 0xda, 0x04, 0x5d, 0xd7,
-			    0x81, 0x9c, 0x7b, 0xc8 },
-	  .offset	= 0x20D8,
-	},
-	{ "Mac OS X.6 MacBook Unibody",
-	  .sha1sum	= { 0xcd, 0x38, 0xe9, 0x08, 0x06, 0xae, 0x06, 0xec,
-			    0xe0, 0x5b, 0x47, 0x26, 0x04, 0x3b, 0xef, 0xec,
-			    0x40, 0x15, 0x73, 0x4d },
-	  .offset	= 0x20d8,
-	},
-	{ "Mac OS X.5.7",
-	  .sha1sum	= { 0x95, 0x6b, 0xf7, 0x0d, 0xa9, 0x03, 0xa0, 0xb4,
-			    0x8b, 0x41, 0x44, 0x87, 0x61, 0xa0, 0xff, 0xd2,
-			    0xd6, 0x93, 0xa2, 0x90 },
-	  .offset	= 0x20D8
-	},
-	{ "Mac OS X.5.6",
-	  .sha1sum      = { 0xa1, 0x7c, 0xfe, 0xe1, 0xaa, 0x48, 0x83, 0xf6,
-			    0x63, 0xb9, 0xf8, 0x95, 0x2d, 0x15, 0x6a, 0xe2,
-			    0x8f, 0x90, 0xd9, 0x11 },
-	  .offset       = 0x20D8
-	},
-	{ "Mac OS X.5.5",
-	  .sha1sum      = { 0x62, 0xe8, 0xed, 0xa7, 0x67, 0xa8, 0x37, 0x4f,
-			    0x1b, 0x13, 0xc4, 0x61, 0xdd, 0x60, 0x8c, 0x84,
-			    0xf3, 0xc8, 0xc0, 0x24 },
-	  .offset       = 0x20D8
-	},
-	{ "Mac OS X.5.x (Macbook Air)",
-	  .sha1sum	= { 0xbd, 0x5d, 0xc3, 0xff, 0xf3, 0xa6, 0x5f, 0x35,
-			    0xa6, 0xb8, 0xa1, 0x81, 0x40, 0x29, 0x8c, 0x06,
-			    0x94, 0xb3, 0xed, 0xc6 },
-	  .offset       = 0x20D8
- 	},
-	{ "Mac OS X.5.5",
-	  .sha1sum	= { 0xf9, 0x91, 0xab, 0x81, 0xf8, 0x98, 0x8f, 0x41,
-			    0x6f, 0x2d, 0xf0, 0x79, 0x2d, 0xf7, 0x60, 0x2c,
-			    0x82, 0x01, 0x2b, 0xeb },
-	  .offset	= 0x20D8
-	},
-	{ "Mac OS X.5.2",
-	  .sha1sum	= { 0x89, 0x84, 0x60, 0x94, 0x2e, 0x12, 0x65, 0xa3, 
-			    0x57, 0xb5, 0xac, 0x86, 0x82, 0x4a, 0xf2, 0xd4,
-			    0xa4, 0xf7, 0x3b, 0x98 },
-	  .offset	= 0x20D8
-	},
-	{ "Mac OS X.5.1 late 2007",
-	  .sha1sum	= { 0x1c, 0x60, 0xef, 0x27, 0xd5, 0x72, 0x21, 0xcf,
-			    0x3d, 0x76, 0x68, 0x7a, 0x49, 0x73, 0xec, 0x72,
-			    0xff, 0x6f, 0xa1, 0x03 },
-	  .offset	= 0x20D8
-	},
-	{ "Mac OS X.5.1 driver 1.0.9",
-	  .sha1sum	= { 0xa1, 0x7b, 0x71, 0xc0, 0xe6, 0x3b, 0xd3, 0x87,
-			    0x82, 0x10, 0x88, 0xdb, 0x3a, 0x66, 0x5b, 0xb8,
-			    0x7a, 0xdc, 0xa2, 0x08 },
-	  .offset	= 0xC404
-	},
-	{ "Mac OS X.5.1",
-	  .sha1sum	= { 0xb6, 0x9f, 0x49, 0xd3, 0xfa, 0x68, 0x58, 0x41,
-			    0x63, 0x24, 0xc3, 0x90, 0xef, 0xfe, 0x14, 0x33,
-			    0x6a, 0x1d, 0xdb, 0x0b },
-	  .offset	= 0xC404
-	},
-	{ "Mac OS X.4 intel",
-	  .sha1sum	= { 0x01, 0xe2, 0x91, 0xd5, 0x29, 0xe7, 0xc1, 0x8d,
-			    0xee, 0xa2, 0xeb, 0xa2, 0x52, 0xd1, 0x81, 0x14,
-			    0xe0, 0x96, 0x27, 0x6e },
-	  .offset	= 0x2060
-	},
-	{ "Mac OS X unknown version",
-	  .sha1sum	= { 0xc6, 0xc9, 0x4d, 0xd7, 0x7b, 0x86, 0x4f, 0x8b,
-			    0x2d, 0x31, 0xab, 0xf3, 0xcb, 0x2d, 0xe4, 0xc9,
-			    0xd1, 0x39, 0xe1, 0xbf },
-	  .offset	= 0x1434
-	},
-	{ "Mac OS X unknown version",
-	  .sha1sum	= { 0xa1, 0x4c, 0x15, 0x9b, 0x17, 0x6d, 0x27, 0xa6,
-			    0xe9, 0x8d, 0xcb, 0x5d, 0xea, 0x5d, 0x78, 0xb8,
-			    0x1e, 0x15, 0xad, 0x41 },
-	  .offset	= 0x23D8
-	},
-	{ "Mac OS X unknown version",
-	  .sha1sum	= { 0x86, 0x43, 0x0c, 0x04, 0xf9, 0xb6, 0x7c, 0x5c,
-			    0x3d, 0x84, 0x40, 0x91, 0x38, 0xa7, 0x67, 0x98,
-			    0x27, 0x02, 0x5e, 0xc2 },
-	  .offset	= 0x1434
-	},
-	{ "Mac OS X.4.10 PPC64",
-	  .sha1sum	= { 0xde, 0xf7, 0xf4, 0xf0, 0x08, 0xd7, 0x8f, 0x39,
-			    0x3f, 0x95, 0x94, 0x6f, 0xbf, 0x85, 0xd7, 0xb1,
-			    0xbd, 0xdc, 0x41, 0x11 },
-	  .offset	= 0x1318
-	},
-};
-
 /*
  * The vanilla firmware have bugs in USB Interface descriptor
  * values. We override buggy value with correct one. Those value are
@@ -177,7 +60,9 @@ struct patch {
 };
 
 /* From linux-2.6:linux/usb/ch9.J */
-#define	USB_CLASS_VIDEO			0x0e
+#define	USB_CLASS_VIDEO			0x0E
+#define USB_CLASS_MISC			0xEF
+#define USB_PROTOCOL_IFACE_ASSOC	0x01
 
 /* From linux-uvc: uvcvideo.h */
 #define SC_VIDEOCONTROL                 0x01
@@ -186,6 +71,28 @@ struct patch {
 #define PC_PROTOCOL_UNDEFINED           0x00
 
 static struct patch patches[] = {
+	{ N_("Fix device descriptor"),
+	  .offset	= 0x2352,
+	  .value	= { USB_CLASS_MISC,
+			    0x02,
+			    USB_PROTOCOL_IFACE_ASSOC }
+	},
+	{ N_("Fix interface assocation descriptor"),
+	  .offset	= 0x2373,
+	  .value	= { 0x08, 0x0b, 0x00 }
+	},
+	{ N_("Fix video interface collection"),
+	  .offset	= 0x2377,
+	  .value	= { USB_CLASS_VIDEO,
+			    SC_VIDEO_INTERFACE_COLLECTION,
+			    PC_PROTOCOL_UNDEFINED }
+	},
+	{ N_("Fix video streaming device qualifier"),
+	  .offset	= 0x2364,
+	  .value	= { USB_CLASS_VIDEO,
+			    SC_VIDEOSTREAMING,
+			    PC_PROTOCOL_UNDEFINED }
+	},
 	{ N_("Fix video control interface descriptor"),
 	  .offset	= 0x2380,
 	  .value	= { USB_CLASS_VIDEO,
@@ -198,13 +105,6 @@ static struct patch patches[] = {
 			    SC_VIDEOSTREAMING,
 			    PC_PROTOCOL_UNDEFINED }
 	},
-	/* needs to be checked */
-	{ N_("Fix video streaming device qualifier"),
-	  .offset	= 0x2364,
-	  .value	= { USB_CLASS_VIDEO,
-			    SC_VIDEOSTREAMING,
-			    PC_PROTOCOL_UNDEFINED }
-	},
 };
 
 /* OPTIONS */
@@ -237,61 +137,68 @@ static GOptionEntry entries[] = {
 
 /* IMPLEMENTATION */
 
-static unsigned char*
-get_sha1sum(char*filename)
+/*
+ * To find the offset in the driver file, use a hexadecimal editor
+ * (e.g. ghex) and search for "USBSend". You will find data like
+ * this :
+ *
+ * "start USBSend #4 returned 0x%x.....X.....4.~.........%\"
+ *                                                  ^
+ *
+ * Each period means 0x00. The firmware start at the ^ sign
+ * position. Read the offset and store in in this array. The tool must
+ * output the sha1sum of the file, use this to fill the .sha1sum
+ * field.
+ */
+static glong
+seek_offset(const gchar* filename)
 {
-	unsigned char *digest;
-	GMappedFile *file;
-	GError *gerr = NULL;
+	gint fd;
+	glong pos;
+	/* This is the preambule of the firmware in the driver file */
+	const guchar needle[8] = {0xC2, 0x34, 0x06, 0x7E,
+				  0x05, 0x00, 0x00, 0x00};
+	guint read_len, chunk_len = 8;
+	guint offset = 8;
+	guchar* chunk = g_new0(guchar, chunk_len);
 
-	digest = g_malloc0 (gcry_md_get_algo_dlen (GCRY_MD_SHA1));
-	if (!(file = g_mapped_file_new (filename, FALSE, &gerr))) {
-		g_error(_("Unable to open driver: %s"),
-			gerr ? gerr->message : "?");
-		return NULL;
+	if ((fd = g_open (filename, O_RDONLY)) == -1) {
+		g_error(_("Unable to open %s."), filename);
+		pos = -1;
+		goto end;
 	}
-	gcry_md_hash_buffer (GCRY_MD_SHA1, digest,
-			     g_mapped_file_get_contents (file),
-			     g_mapped_file_get_length (file));
-	g_mapped_file_free (file);
 
-	return digest;
-}
-
-static gchar*
-sha1sum_string(unsigned char*digest)
-{
-	gchar *sha1sum = "";
-	gint i;
-
-	for (i = 0; i < 20; i++) {
-		sha1sum = g_strdup_printf("%s %02x", sha1sum, digest[i]);
-	}
-	return sha1sum;
-}
-
-static const struct offset*
-find_offset(unsigned char*digest)
-{
-	gint n;
-
-	for (n = 0; n < G_N_ELEMENTS (offsets); n++)
-		if (!memcmp (offsets[n].sha1sum, digest, 20))
+	for (pos = 0x1000; pos < 0xD000; pos++) {
+		if (lseek(fd, pos, SEEK_SET) != pos) {
+			g_error(_("Failed to seek %s at position %x."),
+				filename, (guint) pos);
 			break;
+		}
 
-	if (n == G_N_ELEMENTS (offsets)) {
-		g_warning(_("Unknown driver. Please report it to %s with "
-			    "machine description and Mac OS X version."),
-			  PACKAGE_BUGREPORT);
-		return NULL;
-	} else {
-		/* translators : %s is the known origin of the driver */
-		g_message(_("Found %s driver"),
-			  offsets[n].desc);
-		return &offsets[n];
+		if ((read_len = read(fd, chunk, chunk_len)) != chunk_len) {
+			g_error(_("Failed to read %s at position %x."),
+				filename, (guint) pos);
+			break;
+		}
+
+		if (memcmp(needle, chunk, chunk_len) == 0) {
+			pos+= offset;
+			g_message(_("Found firmware signature at offset 0x%X."),
+				  (guint) pos);
+			goto end_fd;
+		}
 	}
+	pos = -1;
+
+
+ end_fd:
+	close(fd);
+ end:
+	g_free(chunk);
+	return pos;
 }
 
+
 /* extract firmware from filename at offset to output, checking the
    format. */
 static int
@@ -332,9 +239,9 @@ extract(char *filename,
 		}
 		len = (data[0] << 8) | data[1];
 		req = (data[2] << 8) | data[3];
-		if (len == 0x8001)
-			break; /* success */
-		else if (len == 0)
+		if (len == 0x8001) {
+			break;	/* success */
+		} else if (len == 0)
 			continue;
 		else if (len < 0 || len >= 1024) {
 			g_error(_("Invalid firmware data_length %d, "
@@ -359,20 +266,42 @@ extract(char *filename,
 	return 0;
 }
 
+static gint
+firmware_version(int fd, guchar* version)
+{
+	guint read_len;
+
+	if (lseek(fd, 4, SEEK_SET) == -1)
+		g_error(_("Failed to find firmware version, unable to patch it"));
+		
+	if ((read_len = read(fd, version, 3)) != 3)
+		g_error(_("Failed to read firmware version, unable to patch it"));
+
+	g_message(_("Firmware version %d.%d.%d (0x%02X.0x%02X.0x%02X)"),
+		  version[0], version[1], version[2],
+		  version[0], version[1], version[2]);
+
+	return 0;
+}
+
 static int
 patch(char *filename)
 {
 	int fd, i, n, ret = 0;
+	guchar version[3];
 
 	if ((fd = g_open(filename, O_RDWR)) == -1) {
 		g_error(_("Error while opening firmware file for patching."));
 		return -1;
 	}
 
+	firmware_version(fd, version);
+
 	n = G_N_ELEMENTS(patches);
 	for (i = 0; i < n; i++) {
 		g_message(_("Apply patch %i : %s"),
 			  i, gettext(patches[i].desc));
+
 		if (lseek(fd, patches[i].offset, SEEK_SET) == -1) {
 			g_error(_("Unable to patch the firmware."));
 			ret = -1;
@@ -397,11 +326,9 @@ int main (int argc, char *argv[])
 	GOptionContext *context;
 	GError *error = NULL;
 	int status = 0;
-	gchar *sha1sum;
 	gchar *pathname;
-	unsigned char*digest;
-	const struct offset *offset;
-	
+	glong offset;
+
 #if ENABLE_NLS					\
 	/* ??????? */
 	setlocale(LC_ALL, "");
@@ -437,30 +364,30 @@ int main (int argc, char *argv[])
 			error->message); /* howto translate? */
 		fprintf(stderr, g_option_context_get_help(context,
 							  FALSE, NULL));
+		return 1;
 	}
 
-	if (g_access(driver_filename, R_OK)) {
+	if (driver_filename == NULL) {
+	  fprintf(stderr, _("Option -a is mandatory. See --help.\n"));
+	  return 1;
+	}
+
+	if (g_access(driver_filename, R_OK | F_OK)) {
 		g_error(_("Unable to read driver %s."), driver_filename);
+		return 1;
 	}
 
 	pathname = g_build_filename(firmware_dir, firmware_filename, NULL);
 	g_option_context_free(context);
 
-	/* check sha1sum on firmware, to prevent loading crap into the
-	 * iSight and thus possibly damaging it. */
-	digest = get_sha1sum(driver_filename);
-	if (!digest) {
-		return -1;
-	}
+	/* search */
+	offset = seek_offset(driver_filename);
 
-	sha1sum = sha1sum_string(digest);
-	offset = find_offset(digest);
-
-	if (!offset)
+	if (offset <= 0)
 		g_error(_("Unable to find firmware in the file."));
 
 	/* extract */
-	status = extract(driver_filename, offset->offset, pathname);
+	status = extract(driver_filename, offset, pathname);
 	if (!status)
 		g_message(_("Firmware extracted successfully in %s"), pathname);
 
@@ -471,8 +398,5 @@ int main (int argc, char *argv[])
 	else
 		g_error(_("Failed to apply patches to %s"), pathname);
 
-	g_free(digest);
-	g_free(sha1sum);
-
  	return status;
 }
